/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  backend-html.c defines the screen output backend of refdbd
  markus@mhoenicka.de 9-5-00
  $Id: backend-html.c,v 1.27 2004/05/01 23:40:13 mhoenicka Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#include <string.h>
#include <syslog.h> /* for definitions of log message priorities */
#include <dbi/dbi.h>

#include "refdb.h"
#include "linklist.h"
#include "backend.h"
#include "backend-html.h"
#include "writenote.h"
#include "strfncs.h"
#include "refdbd.h"
#include "cgi.h"
#include "dbfncs.h"

extern char refdblib[]; /* location of shareable files */
extern struct BSTRING cgihead_html;
extern struct BSTRING cgihead_plain;
extern struct BSTRING outomem;
extern int n_log_level; /* numeric version of log_level */

/* forward declaration of local functions */
char* print_field_html(const char* item, struct renderinfo* ptr_rendinfo, const char* start_string);

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  prepare_render_html(): writes a header for the html output of a
                         query

  char* prepare_render_html returns a ptr to the string. Due to
                         reallocation this may be different from the
			 ptr passed to this function

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* prepare_render_html(struct renderinfo* ptr_rendinfo) {
  int n_cgi;
  char cgi_buffer[1024];
  char* html_head;
  char* new_ref;
  char* header;

  n_cgi = (ptr_rendinfo->ref_format == REFCGIHTML) ? 1:0;

  /* write a header */
  if (n_cgi) {
    html_head = load_html("refdbq_head", refdblib);
    if (!html_head) {
      return NULL;
    }
    strcpy(*(ptr_rendinfo->ptr_ref), cgihead_html.text);
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), html_head, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      free(html_head);
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
    free(html_head);

    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "<thead><tr><th class=\"result\">Matching References</th></tr>\n<tr><td class=\"result\"><input type=\"submit\" value=\"Delete selected references\" name=\"deleteref\">&nbsp;<input type=\"submit\" value=\"Add selected references to Personal List\" name=\"pickref\">&nbsp;<input type=\"submit\" value=\"Remove selected references from Personal List\" name=\"unpickref\"></td></tr></thead><tbody>", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      free(html_head);
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }
  else { /* not cgi */
    if ((header = assemble_header(ptr_rendinfo)) == NULL) {
      return NULL;
    }

    if ((new_ref = mstrcpy(*(ptr_rendinfo->ptr_ref), header, ptr_rendinfo->ptr_ref_len)) == NULL) {
      free(header);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
      free(header);
    }
  }

  return *(ptr_rendinfo->ptr_ref);
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  finish_render_html(): writes a footer for the html output of a query

  char* finish_render_html returns ptr to the buffer if successful, NULL if failed

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* finish_render_html(struct renderinfo* ptr_rendinfo) {
  int n_cgi;
  char* new_ref;
  char* html_foot;

  n_cgi = (ptr_rendinfo->ref_format == 6) ? 1:0;

  if (n_cgi) {
    html_foot = load_html("refdbq_foot", refdblib);
    if (!html_foot) {
      return NULL;
    }
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), html_foot, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      free(html_foot);
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
    free(html_foot);
    return *(ptr_rendinfo->ptr_ref);
  }
  else {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</body>\n</html>\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
      return new_ref;
    }
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  render_html() renders a RIS dataset for html output

  char* render_html returns ptr to the buffer if successful, NULL if failed

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* render_html(struct renderinfo* ptr_rendinfo) {
  int n_cgi;
  int errcode; /* receives error code for periodical requests */
  int nhave_booktitle = 0;
  int nhave_para = 0;
  char have_page = 0;
  char have_author = 0;
  const char* item;
  const char* item1;
  char id[32] = "";
  char* new_ref;
  const char* type;
  char* avail;
  char *entitize_buf;
  char date_buffer[256];
  char href_buffer[20];
  char hr_buffer[10];
  dbi_result dbires;
  dbi_conn conn;
  struct REPRINT reprint;

  n_cgi = (ptr_rendinfo->ref_format == REFCGIHTML) ? 1:0;

  if (ptr_rendinfo->ref_format == REFXHTML) {
    strcpy(hr_buffer, "<hr />\n");
  }
  else {
    strcpy(hr_buffer, "<hr>\n");
  }

  if ((type = get_refdb_type(ptr_rendinfo->dbires)) == NULL) {
    return NULL;
  }

  conn = dbi_result_get_conn(ptr_rendinfo->dbires);

  if (n_cgi) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "<tr><td class=\"result\">", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }

  /*----------------------------------------------------------------*/
  /* ID */
  get_refdb_id(ptr_rendinfo->dbires, id);

  if (*id) {
    if (n_cgi) {
      /* misuse date_buffer */
      sprintf(date_buffer, "<input type=\"checkbox\" name=\"pickid%s\" value=\"1\">&nbsp;<input type=\"submit\" value=\"Edit\" name=\"editid%s\">", id, id);
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), date_buffer, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }
    /* else: do nothing */
  }
  else {
    return NULL;
  }

  /* each reference is wrapped in a div element */
  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "<div class=\"record\">\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
    LOG_PRINT(LOG_WARNING, outomem.text);
    return NULL;
  }
  else {
    *(ptr_rendinfo->ptr_ref) = new_ref;
  }

  /* misuse date_buffer */
  if (is_in_list(ptr_rendinfo->dbires, ptr_rendinfo->username) == 1) {
    strcpy(date_buffer, "<h2 class='ID'>ID*:");
  }
  else {
    strcpy(date_buffer, "<h2 class='ID'>ID:");
  }

  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), date_buffer, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
    LOG_PRINT(LOG_WARNING, outomem.text);
    return NULL;
  }
  else {
    *(ptr_rendinfo->ptr_ref) = new_ref;
  }

  /* ID */
  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), id, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
    LOG_PRINT(LOG_WARNING, outomem.text);
    return NULL;
  }
  else {
    *(ptr_rendinfo->ptr_ref) = new_ref;
  }

  /*----------------------------------------------------------------*/
  /* pubyear */
  item = get_refdb_pubyear(ptr_rendinfo->dbires, date_buffer);
  if (item != NULL) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), " (", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    /* year */
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), date_buffer, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), ")", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

  }
  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</h2>\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
    LOG_PRINT(LOG_WARNING, outomem.text);
    return NULL;
  }
  else {
    *(ptr_rendinfo->ptr_ref) = new_ref;
  }


  /*----------------------------------------------------------------*/
  /* authors */
  /* request authors */
  dbires = request_authors(conn, 1, NULL, my_dbi_result_get_idval(ptr_rendinfo->dbires, "refdb_id"));
  if (dbires == NULL) {
    return NULL;
  }

  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "<p class='authors'>", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
    clean_request(dbires);
    LOG_PRINT(LOG_WARNING, outomem.text);
    return NULL;
  }
  else {
    *(ptr_rendinfo->ptr_ref) = new_ref;
  }

  /* fetch all authors of this article */
  while ((item = get_author(dbires)) != NULL) {
    have_author = 1;

    if ((entitize_buf = strdup(item)) == NULL
	|| sgml_entitize(&entitize_buf, NULL) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
      
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), entitize_buf, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      clean_request(dbires);
      free(entitize_buf);
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    free(entitize_buf);

    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "; ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      clean_request(dbires);
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }
  clean_request(dbires);

  if (have_author) { /* eliminate the trailing comma after the last author */
    new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</p>\n", ptr_rendinfo->ptr_ref_len, 2);
  }
  else {
    new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</p>\n", ptr_rendinfo->ptr_ref_len, 0);
  }

  if (new_ref == NULL) {
    LOG_PRINT(LOG_WARNING, outomem.text);
    return NULL;
  }
  else {
    *(ptr_rendinfo->ptr_ref) = new_ref;
  }

  /*----------------------------------------------------------------*/
  /* title */
  item = get_refdb_title_copy(ptr_rendinfo->dbires);
  if (item != NULL) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "<p class='title'>", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      free((char*)item);
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    if (sgml_entitize((char**)&item, NULL) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
      
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* title */
      LOG_PRINT(LOG_WARNING, outomem.text);
      free((char*)item);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</p>\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      free((char*)item);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
    free((char*)item);
  }

  if (strcmp(type, "BOOK") == 0 || strcmp(type, "SER") == 0
      || strcmp(type, "THES") == 0 || strcmp(type, "RPRT") == 0
      || strcmp(type, "CHAP") == 0 || strcmp(type, "CONF") == 0
      || strcmp(type, "UNPB") == 0) {

    if (strcmp(type, "CHAP") == 0) {
      /*----------------------------------------------------------------*/
      /* book editors */
      dbires = request_authors(conn, 2, NULL, my_dbi_result_get_idval(ptr_rendinfo->dbires, "refdb_id")); /* book ed. */
      if (dbires == NULL) {
	return NULL;
      }

      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "<p class='chap'>in: ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	clean_request(dbires);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }

      have_author = 0;
      /* fetch all editors of this article */
      while ((item = get_author(dbires)) != NULL) {
	have_author = 1;

	if ((entitize_buf = strdup(item)) != NULL
	    && sgml_entitize(&entitize_buf, NULL) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  return NULL;
	}
      
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), entitize_buf, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  clean_request(dbires);
	  free(entitize_buf);
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}

	free(entitize_buf);

	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "; ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  clean_request(dbires);
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
      }
      clean_request(dbires);

      if (have_author) { /* eliminate the trailing comma after the last author */
	new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "(ed.), ", ptr_rendinfo->ptr_ref_len, 2);
      }
      
      if (new_ref == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }

      /*----------------------------------------------------------------*/
      /* book title */
      item = get_refdb_booktitle_copy(ptr_rendinfo->dbires);
      if (item != NULL) {
	nhave_booktitle++;

	if (sgml_entitize((char**)&item, NULL) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  return NULL;
	}
      
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* book title */
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  free((char*)item);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
	free((char*)item);
      }
    }
    else { /* not chapter */
      /*----------------------------------------------------------------*/
      /* book title */
      item = get_refdb_booktitle_copy(ptr_rendinfo->dbires);
      if (item != NULL) {
	nhave_booktitle++;
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "<p class='book'>", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  free((char*)item);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}

	if (sgml_entitize((char**)&item, NULL) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  return NULL;
	}
      
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* book title */
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  free((char*)item);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
	free((char*)item);
      }
    }

    item1 = get_refdb_title_series_copy(ptr_rendinfo->dbires);
    if (item1 != NULL) {
      dbires = request_authors(conn, 3, NULL, my_dbi_result_get_idval(ptr_rendinfo->dbires, "refdb_id")); /* series ed. */
      if (dbires == NULL) {
	return NULL;
      }
	
      if (nhave_booktitle) {
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "; part of: ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  clean_request(dbires);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
      }
      else {
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "<p>", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  clean_request(dbires);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
      }

      have_author = 0;
      /* fetch all editors of this article */
      while ((item = get_author(dbires)) != NULL) {
	have_author = 1;

	if ((entitize_buf = strdup(item)) != NULL
	    && sgml_entitize(&entitize_buf, NULL) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  return NULL;
	}
      
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), entitize_buf, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  clean_request(dbires);
	  free(entitize_buf);
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}

	free(entitize_buf);

	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "; ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  clean_request(dbires);
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
      }
      clean_request(dbires);
      
      if (have_author) { /* eliminate the trailing comma after the last author */
	new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "(series ed.), ", ptr_rendinfo->ptr_ref_len, 2);
      }
      
      if (new_ref == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      
      if (sgml_entitize((char**)&item1, NULL) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item1, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* series title */
	LOG_PRINT(LOG_WARNING, outomem.text);
	free((char*)item1);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
      free((char*)item1);
    }
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</p>\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }
  else { /* not book et al. */

  /*----------------------------------------------------------------*/
  /* periodical */
    nhave_para = 1;
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "<p>", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    /* fetch journal of this article */
    if ((item = get_periodical(dbi_result_get_conn(ptr_rendinfo->dbires), date_buffer, NULL, 4, &errcode, my_dbi_result_get_idval(ptr_rendinfo->dbires, "refdb_id"))) != NULL) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "<em class='periodical'>", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }

      if ((entitize_buf = strdup(item)) != NULL
	  && sgml_entitize(&entitize_buf, NULL) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), entitize_buf, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	free(entitize_buf);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }

      free(entitize_buf);

      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</em> ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }
  }

  /*----------------------------------------------------------------*/
  /* volume */
  item = get_refdb_volume(ptr_rendinfo->dbires);
  if (item != NULL) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "<em class='volume'>", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    if ((entitize_buf = strdup(item)) != NULL
	&& sgml_entitize(&entitize_buf, NULL) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
      
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), entitize_buf, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* volume */
      LOG_PRINT(LOG_WARNING, outomem.text);
      free(entitize_buf);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
    
    free(entitize_buf);

    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</em>", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }
  
  /*----------------------------------------------------------------*/
  /* issue */
  item = get_refdb_issue(ptr_rendinfo->dbires);  
  if (item != NULL) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "<em class='issue'>(", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    if ((entitize_buf = strdup(item)) != NULL
	&& sgml_entitize(&entitize_buf, NULL) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
      
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), entitize_buf, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* issue */
      LOG_PRINT(LOG_WARNING, outomem.text);
      free(entitize_buf);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    free(entitize_buf);
    
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), ")</em>", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }

  /*----------------------------------------------------------------*/
  /* startpage */
  item = get_refdb_startpage(ptr_rendinfo->dbires);
  if (item != NULL) {
    have_page++;
    if (strcmp(type, "BOOK") == 0 || strcmp(type, "SER") == 0 ||
	strcmp(type, "THES") == 0 || strcmp(type, "RPRT") == 0 ||
	strcmp(type, "CHAP") == 0) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "<em class='page'>pp. ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }
    else {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), ":<em class='page'>", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }

    if ((entitize_buf = strdup(item)) != NULL
	&& sgml_entitize(&entitize_buf, NULL) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
      
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), entitize_buf, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* start page */
      LOG_PRINT(LOG_WARNING, outomem.text);
      free(entitize_buf);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    free(entitize_buf);
  }
  
  /*----------------------------------------------------------------*/
  /* endpage */
  item = get_refdb_endpage(ptr_rendinfo->dbires);
  if (item != NULL) {
    have_page++;
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "-", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    if ((entitize_buf = strdup(item)) != NULL
	&& sgml_entitize(&entitize_buf, NULL) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
      
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), entitize_buf, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* end page */
      LOG_PRINT(LOG_WARNING, outomem.text);
      free(entitize_buf);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    free(entitize_buf);
  }

  if (have_page) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</em>", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }

  if (nhave_para) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</p>\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }
  
  /*----------------------------------------------------------------*/
  /* citation key */
  item = get_refdb_citekey(ptr_rendinfo->dbires);

  /* citation key is preprocessed and cannot contain offending xml chars */
  if (item && *item) {
    if (print_field_html(item, ptr_rendinfo, "<p class='citekey'>CITEKEY:") == NULL) {
      return NULL;
    }
  }

  /*----------------------------------------------------------------*/
  /* abstract */
 if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "N2") != NULL
     || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "AB") != NULL
     || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    item = get_refdb_abstract_copy(ptr_rendinfo->dbires);
    if (item != NULL) {
      if (sgml_entitize((char**)&item, NULL) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	free((char*)item);
	return NULL;
      }
      
      if (print_field_html(item, ptr_rendinfo, "<p class='abstract'>ABSTRACT: ") == NULL) {
	free((char*)item);
	return NULL;
      }
      free((char*)item);
    }
  }

  /*----------------------------------------------------------------*/
  /* address */
 if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "AD") != NULL
     || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    item = get_refdb_address_copy(ptr_rendinfo->dbires);
    if (item != NULL) {
      if (sgml_entitize((char**)&item, NULL) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	free((char*)item);
	return NULL;
      }
      
      if (print_field_html(item, ptr_rendinfo, "<p class='address'>ADDRESS: ") == NULL) {
	free((char*)item);
	return NULL;
      }
      free((char*)item);
    }
  }

  /*----------------------------------------------------------------*/
  /* reprint, availability  */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "RP") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    if (get_reprint(ptr_rendinfo->dbires, &reprint, ptr_rendinfo->username, 4) != NULL && reprint.reprint != NULL && *(reprint.reprint)) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "<p class='rp'>REPRINT: ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }

      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), reprint.reprint, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }

      if (reprint.date != NULL && *(reprint.date)) {
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), " (", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), reprint.date, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), ")", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
      }
    }
    if (reprint.date != NULL && *(reprint.date)) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), ": ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), reprint.date, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }
    if (reprint.avail != NULL && *(reprint.avail)) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), ": ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }

      avail = mstrdup(reprint.avail);
      if (avail == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      if (strncmp(strup(avail), "PATH:", 5) == 0) {
	free(avail);
	/* if pdfroot specifies a protocol, use it as a prefix for
	   the href attribute. If not, add file:// as a pseudo
	   protocol */
	if (strncmp(ptr_rendinfo->pdfroot, "http://", 7) == 0
	    || strncmp(ptr_rendinfo->pdfroot, "HTTP://", 7) == 0
	    || strncmp(ptr_rendinfo->pdfroot, "ftp://", 6) == 0
	    || strncmp(ptr_rendinfo->pdfroot, "FTP://", 6) == 0
	    || strncmp(ptr_rendinfo->pdfroot, "file://", 7) == 0
	    || strncmp(ptr_rendinfo->pdfroot, "FILE://", 7) == 0) {
	  strcpy(href_buffer, "<a href=\"");
	}
	else {
	  strcpy(href_buffer, "<a href=\"file://");
	}

	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), href_buffer, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
	if (*(ptr_rendinfo->pdfroot)) {
	  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), ptr_rendinfo->pdfroot, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	    LOG_PRINT(LOG_WARNING, outomem.text);
	    return NULL;
	  }
	  else {
	    *(ptr_rendinfo->ptr_ref) = new_ref;
	  }
	}

	if ((entitize_buf = strdup(reprint.avail+5)) != NULL
	    && sgml_entitize(&entitize_buf, NULL) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  return NULL;
	}
      
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), entitize_buf, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  free(entitize_buf);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}

	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\">", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  free(entitize_buf);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}

	if (*(ptr_rendinfo->pdfroot)) {
	  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), ptr_rendinfo->pdfroot, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	    LOG_PRINT(LOG_WARNING, outomem.text);
	    return NULL;
	  }
	  else {
	    *(ptr_rendinfo->ptr_ref) = new_ref;
	  }
	}

	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), entitize_buf, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  free(entitize_buf);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}

	free(entitize_buf);

	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</a>", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
      }
      else {
	int n_renderaslink = 0;

	free(avail);
	/* if the availability string starts with a protocol
	   specifier, render the string as a link */
	if (strncmp(reprint.avail, "http://", 7) == 0
	    || strncmp(reprint.avail, "HTTP://", 7) == 0
	    || strncmp(reprint.avail, "ftp://", 6) == 0
	    || strncmp(reprint.avail, "FTP://", 6) == 0
	    || strncmp(reprint.avail, "file://", 7) == 0
	    || strncmp(reprint.avail, "FILE://", 7) == 0) {
	  n_renderaslink++;
	  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "<a href=\"", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	    LOG_PRINT(LOG_WARNING, outomem.text);
	    return NULL;
	  }
	  else {
	    *(ptr_rendinfo->ptr_ref) = new_ref;
	  }

	  if ((entitize_buf = strdup(reprint.avail)) != NULL
	      && sgml_entitize(&entitize_buf, NULL) == NULL) {
	    LOG_PRINT(LOG_WARNING, outomem.text);
	    return NULL;
	  }
      
	  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), entitize_buf, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	    LOG_PRINT(LOG_WARNING, outomem.text);
	    free(entitize_buf);
	    return NULL;
	  }
	  else {
	    *(ptr_rendinfo->ptr_ref) = new_ref;
	  }

	  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\">", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	    LOG_PRINT(LOG_WARNING, outomem.text);
	    free(entitize_buf);
	    return NULL;
	  }
	  else {
	    *(ptr_rendinfo->ptr_ref) = new_ref;
	  }
	}

	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), entitize_buf, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_WARNING, outomem.text);
	  free(entitize_buf);
	  return NULL;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}

	free(entitize_buf);

	if (n_renderaslink) {
	  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</a>", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	    LOG_PRINT(LOG_WARNING, outomem.text);
	    return NULL;
	  }
	  else {
	    *(ptr_rendinfo->ptr_ref) = new_ref;
	  }
	}
      }
    }

    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</p>\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }

  /*----------------------------------------------------------------*/
  /* ISBN/ISSN */
 if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "SN") != NULL
     || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    item = get_refdb_issn(ptr_rendinfo->dbires);

    if (item) {
      if ((entitize_buf = strdup(item)) != NULL
	  && sgml_entitize(&entitize_buf, NULL) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      
      if (print_field_html(entitize_buf, ptr_rendinfo, "<p class='issn'>ISSN: ") == NULL) {
	free(entitize_buf);
	return NULL;
      }

      free(entitize_buf);
    }
  }

  /*----------------------------------------------------------------*/
  /* publisher */
 if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "PB") != NULL
     || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    item = get_refdb_publisher(ptr_rendinfo->dbires);

    if (item) {
      if ((entitize_buf = strdup(item)) != NULL
	  && sgml_entitize(&entitize_buf, NULL) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      
      if (print_field_html(entitize_buf, ptr_rendinfo, "<p class='publisher'>PUBLISHER: ") == NULL) {
	free(entitize_buf);
	return NULL;
      }

      free(entitize_buf);
    }
  }

  /*----------------------------------------------------------------*/
  /* city */
 if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "CY") != NULL
     || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    item = get_refdb_city(ptr_rendinfo->dbires);

    if (item) {
      if ((entitize_buf = strdup(item)) != NULL
	  && sgml_entitize(&entitize_buf, NULL) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      
      if (print_field_html(entitize_buf, ptr_rendinfo, "<p class='city'>CITY: ") == NULL) {
	free(entitize_buf);
	return NULL;
      }

      free(entitize_buf);
    }
  }

  /*----------------------------------------------------------------*/
  /* user fields */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "U1") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    item = get_refdb_user1(ptr_rendinfo->dbires);

    if (item) {
      if ((entitize_buf = strdup(item)) != NULL
	  && sgml_entitize(&entitize_buf, NULL) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      
      if (print_field_html(entitize_buf, ptr_rendinfo, "<p class='u1'>USER1: ") == NULL) {
	free(entitize_buf);
	return NULL;
      }
      
      free(entitize_buf);
    }
  }

  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "U2") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    item = get_refdb_user2(ptr_rendinfo->dbires);
    
    if (item) {
      if ((entitize_buf = strdup(item)) != NULL
	  && sgml_entitize(&entitize_buf, NULL) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      
      if (print_field_html(entitize_buf, ptr_rendinfo, "<p class='u2'>USER2: ") == NULL) {
	free(entitize_buf);
	return NULL;
      }

      free(entitize_buf);
    }
  }

  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "U3") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    item = get_refdb_user3(ptr_rendinfo->dbires);

    if (item) {
      if ((entitize_buf = strdup(item)) != NULL
	  && sgml_entitize(&entitize_buf, NULL) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      
      if (print_field_html(entitize_buf, ptr_rendinfo, "<p class='u3'>USER3: ") == NULL) {
	free(entitize_buf);
	return NULL;
      }

      free(entitize_buf);
    }
  }

  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "U4") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    item = get_refdb_user4(ptr_rendinfo->dbires);

    if (item) {
      if ((entitize_buf = strdup(item)) != NULL
	  && sgml_entitize(&entitize_buf, NULL) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      
      if (print_field_html(entitize_buf, ptr_rendinfo, "<p class='u4'>USER4: ") == NULL) {
	free(entitize_buf);
	return NULL;
      }

      free(entitize_buf);
    }
  }

  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "U5") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    item = get_refdb_user5(ptr_rendinfo->dbires);

    if (item) {
      if ((entitize_buf = strdup(item)) != NULL
	  && sgml_entitize(&entitize_buf, NULL) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      
      if (print_field_html(entitize_buf, ptr_rendinfo, "<p class='u5'>USER5: ") == NULL) {
	free(entitize_buf);
	return NULL;
      }

      free(entitize_buf);
    }
  }

  /*----------------------------------------------------------------*/
  /* misc fields */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "M1") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    item = get_refdb_misc1(ptr_rendinfo->dbires);

    if (item) {
      if ((entitize_buf = strdup(item)) != NULL
	  && sgml_entitize(&entitize_buf, NULL) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      
      if (print_field_html(entitize_buf, ptr_rendinfo, "<p class='m1'>MISC1: ") == NULL) {
	free(entitize_buf);
	return NULL;
      }

      free(entitize_buf);
    }
  }

  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "M2") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    item = get_refdb_misc2(ptr_rendinfo->dbires);

    if (item) {
      if ((entitize_buf = strdup(item)) != NULL
	  && sgml_entitize(&entitize_buf, NULL) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      
      if (print_field_html(entitize_buf, ptr_rendinfo, "<p class='m2'>MISC2: ") == NULL) {
	free(entitize_buf);
	return NULL;
      }

      free(entitize_buf);
    }
  }

  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "M3") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    item = get_refdb_misc3(ptr_rendinfo->dbires);

    if (item) {
      if ((entitize_buf = strdup(item)) != NULL
	  && sgml_entitize(&entitize_buf, NULL) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      
      if (print_field_html(entitize_buf, ptr_rendinfo, "<p class='m3'>MISC3: ") == NULL) {
	free(entitize_buf);
	return NULL;
      }
      
      free(entitize_buf);
    }
  }

  /*----------------------------------------------------------------*/
  /* URL */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "UR") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    item = get_refdb_url_copy(ptr_rendinfo->dbires);
    if (item != NULL) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "<p class='ur'><a href=\"", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }

      if (sgml_entitize((char**)&item, NULL) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	free((char*)item);
	return NULL;
      }
      
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	free((char*)item);
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }

      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\">", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	free((char*)item);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }

      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	free((char*)item);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }

      free((char*)item);

      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</a></p>\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	return NULL;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }
  }

  /*----------------------------------------------------------------*/
  /* notes */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "N1") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    item = get_notes_copy(ptr_rendinfo->dbires, ptr_rendinfo->username);

    if (item != NULL) {
      if (sgml_entitize((char**)&item, NULL) == NULL) {
	LOG_PRINT(LOG_WARNING, outomem.text);
	free((char*)item);
	return NULL;
      }
      
      if (print_field_html(item, ptr_rendinfo, "<p class='note'>NOTES:") == NULL) {
	free((char*)item);
	return NULL;
      }
      free((char*)item);
    }
  }

  /*----------------------------------------------------------------*/
  /* extended notes */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "NX") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    dbi_result orig_dbires;

    /* look for additional notes linked with the record,
       keywords, authors, or periodicals */
    if ((dbires = request_notes_by_ref(dbi_result_get_conn(ptr_rendinfo->dbires), atoll(id), REFERENCE|KEYWORD|AUTHOR|PERIODICAL, ptr_rendinfo->username)) == NULL) {
/*       printf("request_notes_by_ref() failed\n"); */
      free_request_notes_by_ref(dbi_result_get_conn(ptr_rendinfo->dbires));
      return NULL;
    }
    
    /* save original query result. This is legal as dbi_result is a
       pointer in disguise */
    orig_dbires = ptr_rendinfo->dbires;
    ptr_rendinfo->dbires = dbires;

    while (dbi_result_next_row(ptr_rendinfo->dbires) != 0) {
/*       if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\nXNOTE:\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) { */
/* 	LOG_PRINT(LOG_WARNING, outomem.text); */
/* 	return NULL; */
/*       } */
/*       else { */
/* 	*(ptr_rendinfo->ptr_ref) = new_ref; */
/*       } */

      if ((new_ref = render_note(ptr_rendinfo)) == NULL) {
	ptr_rendinfo->dbires = orig_dbires;
	clean_request(dbires);
	free_request_notes_by_ref(dbi_result_get_conn(ptr_rendinfo->dbires));
	return NULL;
      }
    }
    clean_request(dbires);
    free_request_notes_by_ref(dbi_result_get_conn(orig_dbires));
    ptr_rendinfo->dbires = orig_dbires;
  }

  if (n_cgi) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</td></tr>\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }
  else {
    /* finish div */
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</div>\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

/*     if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), hr_buffer, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { */
/*       LOG_PRINT(LOG_WARNING, outomem.text); */
/*       return NULL; */
/*     } */
/*     else { */
/*       *(ptr_rendinfo->ptr_ref) = new_ref; */
/*     } */
  }

/*      printf("%s\n", *(ptr_rendinfo->ptr_ref)); */
  return new_ref;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  print_field_html(): writes simple field contents to HTML output

  char* print_field_html returns a ptr to the modified string. Due to
                         reallocation this may be different from the
			 ptr passed to this function in the ptr_rendinfo
			 structure which is also updated accordingly

  const char* item ptr to the contents of the field to write

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  const char* start_string ptr to a string to be printed before the 
                         field contents

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* print_field_html(const char* item, struct renderinfo* ptr_rendinfo, const char* start_string) {
  char* new_ref;

  if (item != NULL) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)start_string, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</p>\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, outomem.text);
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
      }
  }
  return *(ptr_rendinfo->ptr_ref);
}
