/* refdbd.h header file for refdbd.c */
/* markus@mhoenicka.de 2-10-00 */
/* $Id: refdbd.h.in,v 1.7.2.4 2004/10/25 21:17:34 mhoenicka Exp $ */

/*
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#ifndef XmlParse_INCLUDED
#include <expat.h>
#endif

#ifndef backend_included
#include "backend.h"
#endif

struct ADDRESULT {
  unsigned long long success; /* number of successfully added references */
  unsigned long long failure; /* number of failures */
  unsigned long long updated; /* number of successfully updated references */
  unsigned long long skipped; /* number of skipped references */
  char *msg; /* ptr to optional error message */
  size_t msg_len; /* length of optional error message */
};

struct DELRESULT {
  unsigned long long success; /* number of successfully deleted references */
  unsigned long long failure; /* number of failures */
  unsigned long long skipped; /* number of skipped references */
};

struct CLIENT_REQUEST {
  int fd; /* file descriptor of incoming connection */
  int n_cgi; /* if 1, handle as CGI request */
  int* ptr_optind; /* ptr to index of first parsed argument to handle */
  int inargc; /* number of parsed arguments */
  int protocol; /* protocol version */
  char** inargv; /* ptr to array of parsed arguments */
  char username[USERNAME_LENGTH]; /* username for database authentication */
  char passwd[PREFS_BUF_LEN]; /* password for database authentication */
  char server_ip[PREFS_BUF_LEN]; /* hostname or IP address of database server */
  char my_hostname[256]; /* hostname of box running refdbd */
  char dbs_port_address[PREFS_BUF_LEN]; /* port where database server listens */
  char dbserver[PREFS_BUF_LEN]; /* name of the libdbi driver */
  char current_db[DBNAME_LENGTH+1]; /* name of reference database */
  char db_path[PREFS_BUF_LEN]; /* path of dir containing databases */
  char client_ip[16]; /* IP address of client */
  char* argument; /* additional unparsed command arguments */
};

#define MSG_BUF_SIZE 256
#define FIFO_NAME "/tmp/refdbd_fifo" /* base name of a fifo */
#define MAIN_DB "refdb" /* name of the main database */

#define CITEKEY_BASE_LENGTH 240 /* maximum length of authorname+year */

/* here we define some strings that will be used in the DSSSL style specification files */
#define STYLESHEET_HTML_NAME "docbook-refdb-html.dsl"
#define STYLESHEET_HTML_BASENAME "docbook-refdb-html"
#define STYLESHEET_HTML_ID "html"
#define STYLESHEET_HTML_PUBID "-//Markus Hoenicka//DOCUMENT RefDB DocBook html stylesheet//EN"

#define STYLESHEET_PRINT_NAME "docbook-refdb-print.dsl"
#define STYLESHEET_PRINT_BASENAME "docbook-refdb-print"
#define STYLESHEET_PRINT_ID "print"
#define STYLESHEET_PRINT_PUBID "-//Markus Hoenicka//DOCUMENT RefDB DocBook print stylesheet//EN"

/* these defines are for the bibliography style output */
#define PROCESSING_INST "<?xml version=\"1.0\"?>\n"
#define PROCESSING_INST_UTF8 "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"
#define DOCTYPE_CITESTYLE "<!DOCTYPE CITESTYLE PUBLIC \"-//Markus Hoenicka//DTD CiteStyle V1.3//EN\" \"http://refdb.sourceforge.net/dtd/citestylex-1.3/citestylex.dtd\">\n"
#define DOCTYPE_STYLESET "<!DOCTYPE STYLESET PUBLIC \"-//Markus Hoenicka//DTD CiteStyle V1.3//EN\" \"http://refdb.sourceforge.net/dtd/citestylex-1.3/citestylex.dtd\">\n"

/* this one is for the risx.dtd */
#define RISX_PUBID "<!DOCTYPE ris PUBLIC \"-//Markus Hoenicka//DTD Ris V1.0.2//EN\" \"http://refdb.sourceforge.net/dtd/risx-1.0.2/risx.dtd\" [\n<!ENTITY lt \"&#38;#60;\">\n<!ENTITY gt \"&#62;\">\n<!ENTITY amp \"&#38;#38;\">\n]>\n"

/* the extended notes dtd */
#define XNOTE_PUBID "<!DOCTYPE xnoteset PUBLIC \"-//Markus Hoenicka//DTD Xnote V1.1//EN\" \"http://refdb.sourceforge.net/dtd/xnote-1.1/xnote.dtd\" [\n<!ENTITY lt \"&#38;#60;\">\n<!ENTITY gt \"&#62;\">\n<!ENTITY amp \"&#38;#38;\">\n]>\n"

/* the following defines are for daemonize & friends */
#define ROOT_DIR "/"
#define STDIN_FILENO 0
#define STDOUT_FILENO 1
#define STDERR_FILENO 2

dbi_conn connect_to_db(struct CLIENT_REQUEST* ptr_clrequest, const char* special_db, int nocheck);
void viewstat(struct CLIENT_REQUEST* ptr_clrequest);
void listdb(struct CLIENT_REQUEST* ptr_clrequest, int select);
void listuser(struct CLIENT_REQUEST* ptr_clrequest);
void liststyle(struct CLIENT_REQUEST* ptr_clrequest);
void getstyle(struct CLIENT_REQUEST* ptr_clrequest);
int whichdb(struct CLIENT_REQUEST* ptr_clrequest);
int createdb(struct CLIENT_REQUEST* ptr_clrequest, const char* dbname, const char* db_encoding);
int deletedb(struct CLIENT_REQUEST* ptr_clrequest, const char* dbname);
void child_confserv(struct CLIENT_REQUEST* ptr_clrequest);
void confserv(char *command, char *server_ip);
int addref(struct CLIENT_REQUEST* ptr_clrequest, char* set_owner, struct ADDRESULT* ptr_addresult, int replace_ref, int n_keep_id, const char* reftype, Lilid* ptr_sentinel, char* input_encoding);
int deleteref(struct CLIENT_REQUEST* ptr_clrequest, struct DELRESULT* ptr_delresult);
int delete_ref_by_id(unsigned long long idval, dbi_conn conn, struct CLIENT_REQUEST* ptr_clrequest, struct DELRESULT* ptr_delresult);
unsigned long long getref(struct CLIENT_REQUEST* ptr_clrequest, struct bibinfo* ptr_biblio_info, int ref_format, int n_privatelist, char *pdfroot, char *cgi_url);
int pickref(struct CLIENT_REQUEST* ptr_clrequest, int n_remove, struct DELRESULT* ptr_delresult);
int adduser(struct CLIENT_REQUEST* ptr_clrequest, char *user_host, char* newuser_passwd, int n_remove);
int addword(struct CLIENT_REQUEST* ptr_clrequest, int n_remove);
void getfoo(struct CLIENT_REQUEST* ptr_clrequest, int type, int all, char* input_encoding);
int getbib(struct CLIENT_REQUEST* ptr_clrequest, struct bibinfo* ptr_biblio_info, char *table_name, int n_ref_format, int n_send_stylespec, struct ADDRESULT* ptr_addresult);
char* getid(char* idstring, char** db, int* type);
int addstyle(struct CLIENT_REQUEST* ptr_clrequest, struct ADDRESULT* ptr_addresult);
int deletestyle(struct CLIENT_REQUEST* ptr_clrequest, struct DELRESULT* ptr_delresult);
int remove_style(dbi_conn conn, unsigned long long id);
int read_xml(int fd, XML_Parser p, struct ADDRESULT* ptr_addresult);
void log_print(int priority, const char* string);
int is_refdb_admin(struct CLIENT_REQUEST* ptr_clrequest);
int run_keyword_scan(struct CLIENT_REQUEST* ptr_clrequest, Lilid* ptr_sentinel, int mode);
int addnote(struct CLIENT_REQUEST* ptr_clrequest, char* set_owner, struct ADDRESULT* ptr_addresult, int replace_note, Lilid* ptr_sentinel, char* input_encoding);
int deletenote(struct CLIENT_REQUEST* ptr_clrequest, struct DELRESULT* ptr_delresult);
int deletenote(struct CLIENT_REQUEST* ptr_clrequest, struct DELRESULT* ptr_delresult);
unsigned long long getnote(struct CLIENT_REQUEST* ptr_clrequest, struct bibinfo *ptr_biblio_info, int ref_format, int n_privatelist, char *cgi_url);
int addlink(struct CLIENT_REQUEST* ptr_clrequest, char* set_owner, struct ADDRESULT* ptr_addresult, int n_remove);
struct CLIENT_REQUEST* new_client_request(void);
struct CLIENT_REQUEST* dup_client_request(struct CLIENT_REQUEST* ptr_clrequest);
int initialize_client_request(struct CLIENT_REQUEST* ptr_clrequest);
int free_client_request(struct CLIENT_REQUEST* ptr_clrequest);
int update_meta(dbi_conn conn, struct CLIENT_REQUEST* ptr_clrequest);
int check_is_subseq(const char* author_concat, const char* table_name, int n_currpos, short int n_pubyear, dbi_conn conn);





