/* writenote.c: functions to write notes */
/* markus@mhoenicka.de 2003-10-12 */
/* $Id: writenote.c,v 1.3 2003/11/04 23:42:17 mhoenicka Exp $ */

/*
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <dbi/dbi.h>

#include "backend.h"
#include "backendn-scrn.h"
#include "backendn-notex.h"
#include "backendn-html.h"


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  render_note(): formats the result of a note query

  char* render_note returns ptr to the buffer if successful, NULL if failed

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* render_note(struct renderinfo* ptr_rendinfo) {
  /* ToDo: ptr_rendinfo->ref_format contains the requested note format
     dispatch to appropriate backend */
  char* new_ref;
  
  if (ptr_rendinfo->ref_format == XNOTE) {
    new_ref = render_note_xnote(ptr_rendinfo);
  }
  else if (ptr_rendinfo->ref_format == REFHTML
	   || ptr_rendinfo->ref_format == REFXHTML) {
    new_ref = render_note_html(ptr_rendinfo);
  }
  else {
    new_ref = render_note_scrn(ptr_rendinfo);
  }
  
  if (new_ref) {
    *(ptr_rendinfo->ptr_ref) = new_ref;
    return new_ref;
  }
  else {
    return NULL;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  prepare_render_note(): writes a header for the output of a note query

  char* prepare_render returns a ptr to the modified string if
                       successful, NULL if failed

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* prepare_render_note(struct renderinfo* ptr_rendinfo) {
  char* new_ref;
  
  if (ptr_rendinfo->ref_format == XNOTE) {
    new_ref = prepare_render_note_xnote(ptr_rendinfo);
  }
  else if (ptr_rendinfo->ref_format == REFHTML
	   || ptr_rendinfo->ref_format == REFXHTML) {
    new_ref = prepare_render_note_html(ptr_rendinfo);
  }
  else {
    new_ref = prepare_render_note_scrn(ptr_rendinfo);
  }
  
  if (new_ref) {
    *(ptr_rendinfo->ptr_ref) = new_ref;
    return new_ref;
  }
  else {
    return NULL;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  finish_render_note(): writes a footer for the output of a note query

  char* finish_render_note returns ptr to the buffer if successful,
                           NULL if failed

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* finish_render_note(struct renderinfo* ptr_rendinfo) {
  char* new_ref;
  
  if (ptr_rendinfo->ref_format == XNOTE) {
    new_ref = finish_render_note_xnote(ptr_rendinfo);
  }
  else if (ptr_rendinfo->ref_format == REFHTML
	   || ptr_rendinfo->ref_format == REFXHTML) {
    new_ref = finish_render_note_html(ptr_rendinfo);
  }
  else {
    new_ref = finish_render_note_scrn(ptr_rendinfo);
  }
  
  if (new_ref) {
    *(ptr_rendinfo->ptr_ref) = new_ref;
    return new_ref;
  }
  else {
    return NULL;
  }
}
